/*
Copyright (c) 1993, Clemens Wagner
 
Permission to use, copy, modify, and distribute this software and 
its documentation for any purpose is hereby granted without fee, provided
that
(i) you have alays to distribute the whole archive
(ii) the above copyright notices and this permission notice appear in
all copies of the software and related documentation, changed or new
code has to be marked
(iii) else, there are the same conditions as for InterViews

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 */

#include <Graphic/polygraphic.h>
#include <Graphic/glyphviewer.h>

#include <InterViews/brush.h>
#include <InterViews/canvas.h>
#include <InterViews/color.h>
#include <InterViews/layout.h>
#include <InterViews/polyglyph.h>
#include <InterViews/transformer.h>

PolyGraphic::PolyGraphic(Graphic *gr): Graphic(gr)
{
	LayoutKit *layout=LayoutKit::instance();
	_body=layout->overlay();
}

PolyGraphic::~PolyGraphic()
{
	delete _body;
}

void PolyGraphic::request(Requisition &req) const
{ 
	_body->request(req);
	Requirement &rx=req.x_requirement();
	Requirement &ry=req.y_requirement();
	Coord left,bottom,right,top;

	left=-rx.natural()*rx.alignment();
	right=left+rx.natural();
	bottom=-ry.natural()*ry.alignment();
	top=bottom+ry.natural();

	if(_t!=nil)
		corners(left,bottom,right,top,*_t);
	rx.natural(right-left);
	rx.stretch(0.0);
	rx.shrink(0.0);
	rx.alignment(-left/rx.natural());

	ry.natural(top-bottom);
	ry.stretch(0.0);
	ry.shrink(0.0);
	ry.alignment(-bottom/ry.natural());
}

void PolyGraphic::allocate(Canvas *c,const Allocation &a,Extension &ext)
{
	if(_t!=nil)
	{
		c->push_transform();
		c->transform(*_t);
	} 
	_body->allocate(c,a,ext);
	if(_t!=nil)
		c->pop_transform();
}

void PolyGraphic::undraw()
{
	_body->undraw();
}

void PolyGraphic::append_(Graphic *g)
{ 
	_body->append(g);
	g->parent(this);
}

void PolyGraphic::prepend_(Graphic *g)
{ 
	_body->prepend(g);
	g->parent(this);
}

void PolyGraphic::insert_(GlyphIndex i,Graphic *g)
{ 
	_body->insert(i,g);
	g->parent(this);
}

void PolyGraphic::remove_(GlyphIndex i)
{ 
	Graphic *gr=(Graphic *) _body->component(i);
	if(gr!=nil)
		gr->parent(nil);
	_body->remove(i);
}

void PolyGraphic::replace_(GlyphIndex i,Graphic *g)
{ 
	Graphic *gr=(Graphic *) _body->component(i);
	gr->parent(nil);
	_body->replace(i,g);
	g->parent(this);
}

void PolyGraphic::change_(GlyphIndex i)
{
	_body->change(i);
}

void PolyGraphic::modified_(GlyphIndex i)
{
	_body->modified(i);
}

void PolyGraphic::flush()
{
	GlyphIndex count=count_();

	for(GlyphIndex i=0;i<count;i++)
	{
		Graphic *gr=component_(i);
		concat_(gr,gr,this,gr);
		gr->flush();
	}
	Graphic n;
	*((Graphic *) this)=*(&n);
}

Glyph *PolyGraphic::clone() const
{
	PolyGraphic *pg=new PolyGraphic((Graphic *) this);
	GlyphIndex i,count=_body->count();

	for(i=0;i<count;i++)
	{
		Glyph *gr=_body->component(i);

		pg->append_((Graphic *) gr->clone());
	}
	return pg;
}
	
GlyphIndex PolyGraphic::count_() const
{
	return _body->count();
}

Graphic *PolyGraphic::component_(GlyphIndex i) const
{ 
	return (Graphic *) _body->component(i);
}

Graphic *PolyGraphic::first_containing(PointObj &pt)
{
	GlyphIndex count=_body->count();

	for(GlyphIndex i=0;i<count;i++)
	{
		Graphic *gr=(Graphic *) _body->component(i);

		if(gr->contains(pt))
			return gr;
	}
	return nil;
}

Graphic *PolyGraphic::last_containing(PointObj &pt)
{
	GlyphIndex count=_body->count();

	for(GlyphIndex i=count-1;i >= 0;i--)
	{
		Graphic *gr=(Graphic *) _body->component(i);

		if(gr->contains(pt))
			return gr;
	}
	return nil;
}

Graphic *PolyGraphic::first_intersecting(BoxObj &b)
{
	GlyphIndex count=_body->count();

	for(GlyphIndex i=0;i<count;i++)
	{
		Graphic *gr=(Graphic *) _body->component(i);

		if(gr->intersects(b))
			return gr;
	}
	return nil;
}

Graphic *PolyGraphic::last_intersecting(BoxObj &b)
{
	GlyphIndex count=_body->count();

	for(GlyphIndex i=count-1;i >= 0;i--)
	{
		Graphic *gr=(Graphic *) _body->component(i);

		if(gr->intersects(b))
			return gr;
	}
	return nil;
}

Graphic *PolyGraphic::first_within(BoxObj &gb)
{
	GlyphIndex count=_body->count();
	Coord l,b,r,t;

	for(GlyphIndex i=0;i<count;i++)
	{
		Graphic *gr=(Graphic *) _body->component(i);
		gr->getbounds(l,b,r,t);

		BoxObj box(l,b,r,t);
		if(box.Within(gb))
			return gr;
	}
	return nil;
}

Graphic *PolyGraphic::last_within(BoxObj &gb)
{
	GlyphIndex count=_body->count();
	Coord l,b,r,t;

	for(GlyphIndex i=count-1;i >= 0;i--)
	{
		Graphic *gr=(Graphic *) _body->component(i);
		gr->getbounds(l,b,r,t);

		BoxObj box(l,b,r,t);
		if(box.Within(gb))
			return gr;
	}
	return nil;
}

void PolyGraphic::disconnect_all()
{
	GlyphIndex i,count=count_();

	for(i=0;i<count;i++)
		component_(i)->disconnect_all();
}

/*
void PolyGraphic::update(GlyphViewer *gv)
{
	GlyphIndex i,count=_body->count();

	for(i=0;i<count;i++)
		((Graphic *) _body->component(i))->update(gv);
}
*/

ClassId PolyGraphic::get_classid() const
{
	return POLY_GRAPHIC;
}

boolean PolyGraphic::is_a(ClassId id) const
{
	return id==POLY_GRAPHIC || Graphic::is_a(id);
}

//

void PolyGraphic::draw_gs(Canvas *c,Graphic *gs)
{
	Graphic gstemp;

	GlyphIndex count=_body->count();
	for(GlyphIndex i=0;i<count;i++)
	{
		Graphic *gr=(Graphic *) _body->component(i);
		concat_(gr,gr,gs,&gstemp);
		draw_(gr,c,&gstemp);
	}
	/*
	c->push_transform();
	if(_handles!=nil)
	{
		c->transformer(nil);
		create_handles();
		if(_handles->drawn())
			_handles->redraw();
		else
			_handles->draw(c);
	}
	c->pop_transform();
	*/
}

void PolyGraphic::drawclipped_gs(Canvas *c,Coord l,Coord b,Coord r,Coord t,Graphic *gs)
{
	Graphic gstemp;
	
	GlyphIndex count=_body->count();
	for(GlyphIndex i=0;i<count;i++)
	{
		Graphic *gr=(Graphic *) _body->component(i);
		concat_(gr,gr,gs,&gstemp);
		drawclipped_(gr,c,l,b,r,t,&gstemp);
	}
	/*
	c->push_transform();
	if(_handles!=nil)
	{
		c->transformer(nil);
		create_handles();
		if(_handles->drawn())
			_handles->redraw();
		else
			_handles->draw(c);
	}
	c->pop_transform();
	*/
}

void PolyGraphic::getextent_gs(Coord &l,Coord &b,Coord &cx,Coord &cy,float &tol,Graphic *gs)
{
	Extent e;
	l=b=cx=cy=tol=0.0;
		
	Graphic gstemp;
	Transformer *ttemp=new Transformer;
	Extent te;
	gstemp.transformer(ttemp);

	GlyphIndex count=_body->count();
	for(GlyphIndex i=0;i<count;i++)
	{
		Graphic *gr=(Graphic *) _body->component(i);
		
		concatgs_(gr,gr,gs,&gstemp);
		concatXform_(gr,nil,gr->transformer(),ttemp);
		getextent_(gr,te._left,te._bottom,te._cx,te._cy,te._tol,&gstemp);
		e.Merge(te);
	}
	l=e._left;
	b=e._bottom;
	cx=l+(e._cx-l)*2.0;
	cy=b+(e._cy-b)*2.0;
	tol=e._tol;

	Transformer *tx=gs->transformer();
	if(tx!=nil)
		corners(l,b,cx,cy,*tx);
	cx=(cx+l)/2.0;
	cy=(cy+b)/2.0;
	Resource::unref(ttemp);
}

boolean PolyGraphic::contains_gs(PointObj &po,Graphic *gs)
{
	GlyphIndex count=_body->count();
	Graphic gstemp;
	Transformer ttemp;

	gstemp.transformer(&ttemp);
	for(GlyphIndex i=0;i<count;i++)
	{
		Graphic *gr=(Graphic *) _body->component(i);
		concat_(gr,gr,gs,&gstemp);
		
		if(contains_(gr,po,&gstemp))
		{
			gstemp.transformer(nil);
			return true;
		}
	}
	gstemp.transformer(nil);/*to avoid deleting ttemp explicitly*/
	return false;
}

boolean PolyGraphic::intersects_gs(BoxObj &box,Graphic *gs)
{
	GlyphIndex count=_body->count();
	Graphic gstemp;
	Transformer ttemp;

	gstemp.transformer(&ttemp);
	for(GlyphIndex i=0;i<count;i++)
	{
		Graphic *gr=(Graphic *) _body->component(i);
		concat_(gr,gr,gs,&gstemp);
		
		if(intersects_(gr,box,&gstemp))
		{
			gstemp.transformer(nil);
			return true;
		}
	}
	gstemp.transformer(nil);/*to avoid deleting ttemp explicitly*/
	return false;
}
